package sun.nio.ch;

import java.io.Closeable;
import java.security.AccessController;
import sun.security.action.GetPropertyAction;
import java.nio.channels.ShutdownChannelGroupException;
import java.util.concurrent.RejectedExecutionException;
import java.io.FileDescriptor;
import java.nio.channels.Channel;
import java.util.Iterator;
import java.io.IOException;
import java.util.HashSet;
import java.util.HashMap;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.nio.channels.spi.AsynchronousChannelProvider;
import java.util.Set;
import java.util.Map;
import java.util.concurrent.locks.ReadWriteLock;
import sun.misc.Unsafe;

class Iocp extends AsynchronousChannelGroupImpl
{
    private static final Unsafe unsafe;
    private static final long INVALID_HANDLE_VALUE = -1L;
    private static final boolean supportsThreadAgnosticIo;
    private final ReadWriteLock keyToChannelLock;
    private final Map<Integer, Iocp.OverlappedChannel> keyToChannel;
    private int nextCompletionKey;
    private final long port;
    private boolean closed;
    private final Set<Long> staleIoSet;
    
    Iocp(AsynchronousChannelProvider asynchronousChannelProvider, ThreadPool threadPool) throws IOException {
        super(asynchronousChannelProvider, threadPool);
        this.keyToChannelLock = new ReentrantReadWriteLock();
        this.keyToChannel = new HashMap();
        this.staleIoSet = new HashSet();
        this.port = createIoCompletionPort(-1L, 0L, 0, this.fixedThreadCount());
        this.nextCompletionKey = 1;
    }
    
    Iocp start() {
        this.startThreads(new Iocp.EventHandlerTask());
        return this;
    }
    
    static boolean supportsThreadAgnosticIo() {
        return Iocp.supportsThreadAgnosticIo;
    }
    
    void implClose() {
        synchronized (this) {
            if (this.closed) {
                return;
            }
            this.closed = true;
        }
        close0(this.port);
        synchronized (this.staleIoSet) {
            Iterator<Long> iterator = this.staleIoSet.iterator();
            while (iterator.hasNext()) {
                Iocp.unsafe.freeMemory((long)iterator.next());
            }
            this.staleIoSet.clear();
        }
    }
    
    @Override
    boolean isEmpty() {
        this.keyToChannelLock.writeLock().lock();
        try {
            return this.keyToChannel.isEmpty();
        }
        finally {
            this.keyToChannelLock.writeLock().unlock();
        }
    }
    
    @Override
    final Object attachForeignChannel(final Channel channel, FileDescriptor fileDescriptor) throws IOException {
        return Integer.valueOf(this.associate(new Iocp.OverlappedChannel() {
            @Override
            public <V, A> PendingFuture<V, A> getByOverlapped(long n) {
                return null;
            }
            
            @Override
            public void close() throws IOException {
                channel.close();
            }
        }, 0L));
    }
    
    @Override
    final void detachForeignChannel(Object o) {
        this.disassociate((int)o);
    }
    
    @Override
    void closeAllChannels() {
        Iocp.OverlappedChannel[] array = new Iocp.OverlappedChannel[32];
        int i;
        do {
            this.keyToChannelLock.writeLock().lock();
            i = 0;
            try {
                Iterator<Integer> iterator = this.keyToChannel.keySet().iterator();
                while (iterator.hasNext()) {
                    array[i++] = (Iocp.OverlappedChannel)this.keyToChannel.get((Object)iterator.next());
                    if (i >= 32) {
                        break;
                    }
                    continue;
                }
            }
            finally {
                this.keyToChannelLock.writeLock().unlock();
            }
            for (int j = 0; j < i; ++j) {
                try {
                    array[j].close();
                }
                catch (IOException ex) {}
            }
        } while (i > 0);
    }
    
    private void wakeup() {
        try {
            postQueuedCompletionStatus(this.port, 0);
        }
        catch (IOException detailMessage) {
            throw new AssertionError(detailMessage);
        }
    }
    
    @Override
    void executeOnHandlerTask(Runnable runnable) {
        synchronized (this) {
            if (this.closed) {
                throw new RejectedExecutionException();
            }
            this.offerTask(runnable);
            this.wakeup();
        }
    }
    
    @Override
    void shutdownHandlerTasks() {
        int threadCount = this.threadCount();
        while (threadCount-- > 0) {
            this.wakeup();
        }
    }
    
    int associate(Iocp.OverlappedChannel overlappedChannel, long n) throws IOException {
        this.keyToChannelLock.writeLock().lock();
        int n2;
        try {
            if (this.isShutdown()) {
                throw new ShutdownChannelGroupException();
            }
            do {
                n2 = this.nextCompletionKey++;
            } while (n2 == 0 || this.keyToChannel.containsKey(Integer.valueOf(n2)));
            if (n != 0L) {
                createIoCompletionPort(n, this.port, n2, 0);
            }
            this.keyToChannel.put(Integer.valueOf(n2), overlappedChannel);
        }
        finally {
            this.keyToChannelLock.writeLock().unlock();
        }
        return n2;
    }
    
    void disassociate(int i) {
        boolean n = false;
        this.keyToChannelLock.writeLock().lock();
        try {
            this.keyToChannel.remove(Integer.valueOf(i));
            if (this.keyToChannel.isEmpty()) {
                n = true;
            }
        }
        finally {
            this.keyToChannelLock.writeLock().unlock();
        }
        if (n && this.isShutdown()) {
            try {
                this.shutdownNow();
            }
            catch (IOException ex) {}
        }
    }
    
    void makeStale(Long n) {
        synchronized (this.staleIoSet) {
            this.staleIoSet.add(n);
        }
    }
    
    private void checkIfStale(long n) {
        synchronized (this.staleIoSet) {
            if (this.staleIoSet.remove(Long.valueOf(n))) {
                Iocp.unsafe.freeMemory(n);
            }
        }
    }
    
    private static IOException translateErrorToIOException(int i) {
        String message = getErrorMessage(i);
        if (message == null) {
            message = "Unknown error: 0x0" + Integer.toHexString(i);
        }
        return new IOException(message);
    }
    
    private static native void initIDs();
    
    private static native long createIoCompletionPort(long p0, long p1, int p2, int p3) throws IOException;
    
    private static native void close0(long p0);
    
    private static native void getQueuedCompletionStatus(long p0, Iocp.CompletionStatus p1) throws IOException;
    
    private static native void postQueuedCompletionStatus(long p0, int p1) throws IOException;
    
    private static native String getErrorMessage(int p0);
    
    static void access$300(long n, Iocp.CompletionStatus completionStatus) throws IOException {
        getQueuedCompletionStatus(n, completionStatus);
    }
    
    static void access$600(Iocp iocp, long n) {
        iocp.checkIfStale(n);
    }
    
    static IOException access$700(int n) {
        return translateErrorToIOException(n);
    }
    
    static {
        unsafe = Unsafe.getUnsafe();
        IOUtil.load();
        initIDs();
        supportsThreadAgnosticIo = (Integer.parseInt(((String)AccessController.doPrivileged(new GetPropertyAction("os.version"))).split("\\.")[0]) >= 6);
    }
    
    private class EventHandlerTask implements Runnable
    {
        @Override
        public void run() {
            Invoker.GroupAndInvokeCount groupAndInvokeCount = Invoker.getGroupAndInvokeCount();
            int n = groupAndInvokeCount != null;
            Iocp.CompletionStatus completionStatus = new Iocp.CompletionStatus();
            boolean n2 = false;
            while (true) {
                try {
                    if (groupAndInvokeCount != null) {
                        groupAndInvokeCount.resetInvokeCount();
                    }
                    n2 = false;
                    try {
                        Iocp.access$300(Iocp.this.port, completionStatus);
                    }
                    catch (IOException ex) {
                        ex.printStackTrace();
                        return;
                    }
                    if (completionStatus.completionKey() == 0 && completionStatus.overlapped() == 0L) {
                        Runnable pollTask = Iocp.this.pollTask();
                        if (pollTask != null) {
                            n2 = true;
                            pollTask.run();
                            continue;
                        }
                    }
                    else {
                        Iocp.OverlappedChannel overlappedChannel = null;
                        Iocp.this.keyToChannelLock.readLock().lock();
                        try {
                            overlappedChannel = (Iocp.OverlappedChannel)Iocp.this.keyToChannel.get(Integer.valueOf(completionStatus.completionKey()));
                            if (overlappedChannel == null) {
                                Iocp.access$600(Iocp.this, completionStatus.overlapped());
                                continue;
                            }
                        }
                        finally {
                            Iocp.this.keyToChannelLock.readLock().unlock();
                        }
                        PendingFuture<Object, Object> byOverlapped = overlappedChannel.getByOverlapped(completionStatus.overlapped());
                        if (byOverlapped == null) {
                            Iocp.access$600(Iocp.this, completionStatus.overlapped());
                            continue;
                        }
                        synchronized (byOverlapped) {
                            if (byOverlapped.isDone()) {
                                continue;
                            }
                        }
                        int error = completionStatus.error();
                        Iocp.ResultHandler resultHandler = (Iocp.ResultHandler)byOverlapped.getContext();
                        n2 = true;
                        if (error == 0) {
                            resultHandler.completed(completionStatus.bytesTransferred(), (boolean)n);
                        }
                        else {
                            resultHandler.failed(error, Iocp.access$700(error));
                        }
                        continue;
                    }
                    return;
                }
                finally {
                    if (Iocp.this.threadExit(this, (boolean)n2) == 0 && Iocp.this.isShutdown()) {
                        Iocp.this.implClose();
                    }
                }
            }
        }
    }
    
    private static class CompletionStatus
    {
        private int error;
        private int bytesTransferred;
        private int completionKey;
        private long overlapped;
        
        int error() {
            return this.error;
        }
        
        int bytesTransferred() {
            return this.bytesTransferred;
        }
        
        int completionKey() {
            return this.completionKey;
        }
        
        long overlapped() {
            return this.overlapped;
        }
    }
    
    interface OverlappedChannel extends Closeable
    {
        public abstract <V, A> PendingFuture<V, A> getByOverlapped(long p0);
    }
    
    interface ResultHandler
    {
        public abstract void completed(int p0, boolean p1);
        
        public abstract void failed(int p0, IOException p1);
    }
}
